/**
 * Util.java
 * Created on 2005/05/16
 * Copyright (c) by es , All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND 
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF 
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package izayoi;

import izayoi.cache.Record;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.text.SimpleDateFormat;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import javax.servlet.ServletContext;

public class Util {

    private static ServletContext ctx;
    private static final MessageDigest md5;
    private static final Map TypeExecConstants = new HashMap();
    private static final Map TypeNameConstants = new HashMap();
    public static final Charset utf8Charset = Charset.forName("UTF-8");

    static {
        try {
            md5 = MessageDigest.getInstance("MD5");
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException(e);
        }

        TypeNameConstants.put(Constants.TYPE_LIST, "Xg");
        TypeNameConstants.put(Constants.TYPE_THREAD, "Xbh");
        TypeNameConstants.put(Constants.TYPE_NOTE, "m[g");

        //@TODO x^ł߂
        TypeExecConstants.put(Constants.TYPE_LIST, "list.cgi");
        TypeExecConstants.put(Constants.TYPE_THREAD, "thread.cgi");
        TypeExecConstants.put(Constants.TYPE_NOTE, "note.cgi");
    }

    /**
     * MD5l
     * @param value vZΏۂ̒l
     * @return MD5l
     */
    public static String calcMD5(String value) {
        if (value == null) {
            return null;
        }

        MessageDigest digest = null;
        try {
            digest = (MessageDigest) md5.clone();
        } catch (CloneNotSupportedException e) {
            //W̃voC_gĂȂoȂ
            throw new RuntimeException(e);
        }

        digest.reset();
        ByteBuffer bytes = utf8Charset.encode(value);
        while (bytes.hasRemaining()) {
            digest.update(bytes.get());
        }
        bytes = ByteBuffer.wrap(digest.digest());
        return parseHexString(bytes).toLowerCase();
    }

    /**
     * tB[h
     * @param name tB[h
     * @return tB[hƂĐȂ^
     */
    public static boolean checkFieldName(String name) {
        if (name == null) {
            return false;
        }

        char c;
        for (int i = 0; i < name.length(); i++) {
            c = name.charAt(i);
            if (c == 0x5f || (0x30 <= c && c <= 0x39) || (0x41 <= c && c <= 0x5a) || (0x61 <= c && c <= 0x7a)) {
                continue;
            }
            return false;
        }
        return true;
    }

    /**
     * xss΍
     * @param str
     * @return
     */
    public static String escape(String str) {
        if (str == null) {
            return "";
        }
        str = str.replaceAll("<", "&lt;");
        str = str.replaceAll(">", "&gt;");
        str = str.replaceAll("\"", "&quot;");
        str = str.replaceAll("&#", "&amp;#");
        str = str.replaceAll("\\r?\\n", "<br>");
        return str;
    }

    public static String getEncode() {
        //ʓ|Ȃ̂őSUTF-8
        return utf8Charset.displayName();
    }

    /**
     * t@CɍvMIME^Cv擾
     * @param filename
     * @return
     */
    public static String getMimeType(String filename) {
        if (ctx != null && filename != null) {
            return ctx.getMimeType(filename);
        } else {
            return null;
        }
    }

    /**
     * kډ摜̗L𔻒f
     * @param timeStamp
     * @param id
     * @return kډ摜̃t@C(Ȃnull)
     */
    public static String getThumbNailPath(Configures conf, Record r) {
        //@TODO łłutilɓ˂ނ͔̂@ȕ
        //@TODO 
        String path = conf.getParameter(Izayoi.CONFIG_MAKE_THUMB_PATH);
        String filename = r.getTimeStamp() + "_" + r.getID() + "." + conf.getParameter(Izayoi.CONFIG_MAKE_THUMB_FORMAT);
        File f = new File(path, filename);
        if (f.exists()) {
            return f.getName();
        } else {
            return null;
        }
    }

    /**
     * t@C^CṽAvP[V擾
     * @param type
     * @return
     */
    public static String getTypeExecutor(String type) {
        return (String) TypeExecConstants.get(type);
    }

    /**
     * t@C^Cv̕\L̎擾
     * @param string
     * @return
     */
    public static String getTypeName(String string) {
        return (String) TypeNameConstants.get(string);
    }

    /**
     * t@C^Cv̕\Lꗗ擾
     * @param string
     * @return
     */
    public static Map getTypeNameMap() {
        return Collections.unmodifiableMap(TypeNameConstants);
    }

    public static String parseByteString(long size) {
        String val;
        if (size < 1024) {
            val = size + "byte";
        } else if (size < 1024 * 1024) {
            val = ((int) (size / 1024f * 1000f)) / 1000f + "KB";
        } else {
            val = ((int) (size / (1024f * 1024f) * 1000f)) / 1000f + "MB";
        }
        return val;
    }

    private static String parseHexString(ByteBuffer b) {
        StringBuffer sb = new StringBuffer();
        while (b.hasRemaining()) {
            int i = 0x000000FF & b.get();
            if (i < 16) {
                sb.append('0');
            }
            sb.append(Integer.toString(i, 16));
        }
        return sb.toString();
    }

    /**
     * w蕶UTF-8ł16i\L
     * @param str ϊΏ
     * @return 
     */
    public static String parseHexString(String str) {
        if (str == null) {
            return null;
        }

        ByteBuffer bytes = utf8Charset.encode(str);
        return parseHexString(bytes).toUpperCase();
    }

    /**
     * UTF-8ł16i\L𕜍
     * @param hexString ϊΏ
     * @return 
     */
    public static String parseString(String hexString) {
        if (hexString == null) {
            return null;
        }

        ByteArrayOutputStream bytes = new ByteArrayOutputStream();
        for (int i = 0; i < hexString.length(); i += 2) {
            try {
                bytes.write(Integer.parseInt(hexString.substring(i, i + 2), 16));
            } catch (Exception e) {
                throw new IllegalArgumentException("wrong hexString");
            }
        }
        ByteBuffer b = ByteBuffer.wrap(bytes.toByteArray());
        return utf8Charset.decode(b).toString();
    }

    /**
     * getMimeTypepɃ}bsOT[ubgReiq(ʓ|)
     * @param context
     */
    synchronized public static final void setServletContext(ServletContext context) {
        //@TODO ƃL
        if (ctx == null && context != null) {
            ctx = context;
        }
    }

    /**
     * ̕URLfR[h
     * @param string Ώ
     * @return 
     */
    public static String urlDecode(String string) {
        if (string == null) {
            return null;
        }

        String s = null;
        try {
            s = URLDecoder.decode(string, "UTF-8");
        } catch (UnsupportedEncodingException e) {
            //UTF-8ŗO͏oȂ
            throw new RuntimeException(e);
        }
        return s;
    }

    /**
     * ̕URLGR[h
     * @param string Ώ
     * @return 
     */
    public static String urlEncode(String string) {
        if (string == null) {
            return null;
        }

        String s = null;
        try {
            s = URLEncoder.encode(string, "UTF-8");
        } catch (UnsupportedEncodingException e) {
            //UTF-8ŗO͏oȂ
            throw new RuntimeException(e);
        }
        return s;
    }

    /**
     * @param l
     * @return
     */
    public static String parseDateString(long time) {
		if (time < 9999999999l) {
			time *= 1000;
		}
        return SimpleDateFormat.getDateTimeInstance().format(new Date(time));
    }

    /**
     * @param type
     * @param basename
     * @return
     */
    public static String getFileName(String type, String basename) {
        return type + "_" + parseHexString(basename);
    }
}
