/**
 * CacheEntry.java
 * Created on 2005/05/12
 * Copyright (c) by es , All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND 
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF 
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package izayoi.cache;

import izayoi.Util;

import java.io.IOException;
import java.util.Arrays;
import java.util.List;

/**
 * uLbVv
 * alternate of net.sourceforge.shingetsu.cache.Cache
 * @author es
 */
public abstract class CacheEntry {

    protected final CacheFactory _factory;

    /**
     * @param factory ̃Gg̊Ǘt@Ng
     */
    public CacheEntry(CacheFactory factory) {
        _factory = factory;
    }

    /**
     * GgɃR[hǉ
     * @param record ǉ郌R[h
     * @throws IOException o͗O
     */
    abstract public void append(Record record) throws IOException;

    /**
     * GgɃR[hǉ
     * @param record ǉ郌R[h
     * @throws IOException
     */
    public void append(Record[] records) throws IOException {
        if(records == null) {
            return;
        }
        for (int i = 0; i < records.length; i++) {
            if (records[i].isDeletedRecord()) {
                deleteRecord(records[i]);
            } else {
                append(records[i]);
            }
        }
    }

    /**
     * ̃Gg̏󋵂i
     * @throws IOException o͗O
     */
    abstract public void commit() throws IOException;

    /**
     * Ggw肳ꂽR[h܂ނ?
     * @param record ΏۃR[h
     * @return ܂ނȂ^
     * @throws IOException o͗O
     */
    abstract public boolean contains(Record record) throws IOException;

    /**
     * Ggw肳ꂽR[h폜 ݂Ȃꍇ͉Ȃ<br>
     * ۂ̍폜͎̋Ɉ˂A܂폜ɍ폜ʒmSm[hɗ
     * @param record ΏۃR[h
     * @throws IOException o͗O
     */
    public final void delete(Record record) throws IOException {
        delete(record, true);
    }

    /**
     * Ggw肳ꂽR[h폜 ݂Ȃꍇ͉Ȃ<br>
     * ۂ̍폜͎̋Ɉ˂
     * @param record ΏۃR[h
     * @param isNotine 폜ʒm𑼃m[hɗ?
     * @throws IOException o͗O
     */
    public void delete(Record record, boolean isNotine) throws IOException {
        if (!record.isDeletedRecord() && !contains(record)) {
            return;
        }
        Record delInfo;
        synchronized (this) {
            delInfo = deleteRecord(record);
            append(delInfo);
        }
        if (!isNotine) {
            return;
        }
        noticeMe(delInfo);
    }

    /**
     * Ggw肳ꂽR[h폜 ݂Ȃꍇ͉Ȃ<br>
     * pł̍폜̋Ɏ
     * @param record ΏۃR[h
     * @return 폜R[hɑ΂폜˗R[h
     * @throws IOException o͗O
     */
    abstract protected Record deleteRecord(Record record) throws IOException;

    /**
     * ̃Gg̃x[Xl[擾
     * @return x[Xl[(eLXg\L)
     */
    abstract public String getBaseName();

    /**
     * GgɊ܂܂郌R[h擾
     * @return@݂̃GgɊ܂܂郌R[h
     * @throws IOException o͗O
     */
    abstract public int getCount() throws IOException;

    /**
     * ̃Gg̃GR[hσt@Cl[擾
     * @return t@Cl[
     */
    public final String getEncodedFileName() {
        return Util.urlEncode(getFileName());
    }

    /**
     * ̃Gg̐Factory擾
     * @return Factory
     */
    public final CacheFactory getFactory() {
        return _factory;
    }

    /**
     * ̃Gg̃t@Cl[擾
     * @return t@Cl[
     */
    abstract public String getFileName();

    /**
     * ŏIXV擾
     * @return
     * @throws IOException
     */
    abstract public long getLastModified() throws IOException;

    /**
     * GgɊ܂܂Ō̃R[h擾
     * (W̎ƌ̂œKXI[o[Chׂ)
     * @return@݂̃GgɊ܂܂Ō̃R[h(݂Ȃꍇnull)
     * @throws IOException o͗O
     */
    public Record getLastRecord() throws IOException {
        Record[] r = getRecords(new RecordSorter.ByTimestamp());
        return r.length > 0 ? r[r.length - 1] : null;
    }

    /**
     * w肳ꂽ̃R[h擾
     * @param timestamp@longl
     * @return@ΏۃR[h(݂Ȃꍇnull)
     * @throws IOException o͗O
     */
    public Record getRecord(long timestamp) throws IOException {
        Record[] r = getRecords(Long.toString(timestamp));
        return r.length == 0 ? null : r[0];
    }

    /**
     * w肳ꂽAʎq̃R[h擾
     * @param timestamp@longl
     * @param id ʎq
     * @return@ΏۃR[h(݂Ȃꍇnull)
     * @throws IOException o͗O
     */
    public Record getRecord(long timestamp, String id) throws IOException {
        Record[] r = getRecords(timestamp + "/" + id);
        return r.length == 0 ? null : r[0];
    }

    /**
     * R[h擾
     * @return
     * @throws IOException
     */
    abstract public int getRecordCount() throws IOException;

    /**
     * GgɊ܂܂SR[h擾
     * @return@݂̃GgɊ܂܂SR[h(݂Ȃꍇ͗vf0̔z)
     * @throws IOException o͗O
     */
    abstract public Record[] getRecords() throws IOException;

    /**
     * GgɊ܂܂郌R[h擾
     * (W̎ƌ̂œKXI[o[Chׂ)
     * @param limit ő擾
     * @return@݂̃GgɊ܂܂SR[h(݂Ȃꍇ͗vf0̔z)
     * @throws IOException o͗O
     */
    public Record[] getRecords(int limit) throws IOException {
        return getRecords(limit, null);
    }

    /**
     * GgɊ܂܂郌R[h擾
     * (W̎ƌ̂œKXI[o[Chׂ)
     * @param limit ő擾
     * @param sorter \[gNX (nullŃ\[gȂ)
     * @return@݂̃GgɊ܂܂SR[h(݂Ȃꍇ͗vf0̔z)
     * @throws IOException o͗O
     */
    public Record[] getRecords(int limit, RecordSorter sorter) throws IOException {
        if (limit < 0) {
            limit = 0;
        }
        Record[] r = getRecords(sorter);
        if (r.length > limit) {
            List l = Arrays.asList(r).subList(0, limit);
            r = new Record[l.size()];
            r = (Record[]) l.toArray(r);
        }
        return r;
    }

    /**
     * w肳ꂽ͈͓̎̑SR[h擾
     * @param from longl
     * @param to longl
     * @return@ΏۃR[h(݂Ȃꍇ͗vf0̔z)
     * @throws IOException o͗O
     */
    public Record[] getRecords(long from, long to) throws IOException {
        return getRecords(from + "-" + to);
    }

    /**
     * w肳ꂽ͈͓̎̑SR[h擾
     * @param from longl
     * @param to longl
     * @param sorter \[gNX (nullŃ\[gȂ)
     * @return@ΏۃR[h(݂Ȃꍇ͗vf0̔z)
     * @throws IOException o͗O
     */
    public final Record[] getRecords(long from, long to, RecordSorter sorter) throws IOException {
        Record[] r = getRecords(from, to);
        if (sorter != null) {
            Arrays.sort(r, sorter);
        }
        return r;
    }

    /**
     * GgɊ܂܂SR[h擾
     * @param sorter \[gNX (nullŃ\[gȂ)
     * @return@݂̃GgɊ܂܂SR[h(݂Ȃꍇ͗vf0̔z)
     * @throws IOException o͗O
     */
    public final Record[] getRecords(RecordSorter sorter) throws IOException {
        Record[] r = getRecords();
        if (sorter != null) {
            Arrays.sort(r, sorter);
        }
        return r;
    }

    /**
     * w肳ꂽ\LɍvSR[h擾
     * @param argument \L
     * @return@ΏۃR[h(݂Ȃꍇ͗vf0̔z)
     * @throws IOException o͗O
     */
    abstract public Record[] getRecords(String argument) throws IOException;

    /**
     * w肳ꂽ\LɍvSR[h擾
     * @param argument \L
     * @param sorter \[gNX (nullŃ\[gȂ)
     * @return@ΏۃR[h(݂Ȃꍇ͗vf0̔z)
     * @throws IOException o͗O
     */
    public final Record[] getRecords(String argument, RecordSorter sorter) throws IOException {
        Record[] r = getRecords(argument);
        if (sorter != null) {
            Arrays.sort(r, sorter);
        }
        return r;
    }

    /**
     * Gg̃TCY擾
     * @return Gg̃TCY
     */
    abstract public long getSize() throws IOException;

    /**
     * ̃Gg̎ʂ擾
     * @return 
     */
    abstract public String getType();

    /**
     * w肳ꂽ\LɍvSwb_擾
     * @param argument \L
     * @return@Ώۃwb_(݂Ȃꍇ͗vf0̔z)
     * @throws IOException o͗O
     */
    abstract public String[] head(String argument) throws IOException;

    /**
     * ݂AutoCommit̐ݒl擾
     * @return
     */
    abstract public boolean isAutoCommit();

    /**
     * m[hɍXVʒm
     * @param record XVꂽR[h
     */
    public final void noticeMe(Record record) throws IOException {
        getFactory().noticeUpdate(this, record);
    }

    /**
     * w肳ꂽʎqɍvSR[h擾
     * @param string
     * @return
     */
    abstract public Record getRecord(String id) throws IOException;

    /**
     * append(), delete()ɎIcommit()ĂԂ?
     * @param b Ȃ^
     */
    abstract public void setAutoCommit(boolean b);

    /**
     * ̃Gg̃R[hׂč폜
     * @throws IOException
     */
    abstract public void purge() throws IOException;

}
