/**
 * SyncAgent.java
 * Created on 2005/06/22
 * Copyright (c) by es , All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND 
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF 
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package izayoi.client.agent;

import izayoi.cache.CacheEntry;
import izayoi.cache.GenericRecord;
import izayoi.cache.Record;
import izayoi.client.AbstractAgent;
import izayoi.client.Client;
import izayoi.node.AsynchronousNode;
import izayoi.node.AsynchronousNodeListener;
import izayoi.node.Node;
import izayoi.node.TalkingEvent;

import java.io.IOException;
import java.net.ProtocolException;

/**
 * ̃m[hLbV̎m[hł̍ŏIXVȍ~̃R[h擾Agent
 * @author es
 */
public class SyncAgent extends AbstractAgent implements AsynchronousNodeListener {

    private String id;
    private long timeStamp;

    /**
     * @param boss
     * @param target
     */
    public SyncAgent(Client boss, Node target, CacheEntry cache) {
        super(boss, target, cache);
    }

    /* ( Javadoc)
     * @see izayoi.node.AsynchronousNodeListener#commFinished(izayoi.node.TalkingEvent)
     */
    public void commFinished(TalkingEvent event) {
        try {
            if (event.getResult()) {
                storeRecord(GenericRecord.parseRecord(event.getResponse()));
            } else {
                event.getNegateDetail().printStackTrace(System.err);
            }
        } finally {
            event.getSourceNode().removeNodeListener(this);
        }
    }

    /**
     * @param time
     */
    private void getRecord(String time) throws IOException {
        if (target instanceof AsynchronousNode) {
            //񓯊\Ȃ񓯊
            AsynchronousNode n = (AsynchronousNode) target;
            n.addNodeListener(this);
            n.asyncGet(cache.getFileName(), time);
        } else {
            Record[] r = null;
            try {
                r = target.get(cache.getFileName(), time);
            } catch (ProtocolException e) {
                e.printStackTrace(System.err);
            }
            storeRecord(r);
        }
    }

    /* ( Javadoc)
     * @see izayoi.client.Agent#invoke()
     */
    public void invoke() throws IOException {
        long last = cache.getLastModified();
        String time = null;
        if (last < 0) {
            time = "0-";
        } else {
            if (last > timeStamp) {
                time = timeStamp + "/" + id;
            } else {
                time = last + "-";
            }
        }
        getRecord(time);
    }

    /**
     * @param string
     */
    public void setID(String string) {
        id = string;
    }

    /**
     * @param l
     */
    public void setTimeStamp(long l) {
        timeStamp = l;
    }

    private void storeRecord(Record[] r) {
        try {
            cache.append(r);
        } catch (IOException e) {
            e.printStackTrace(System.err);
        }
    }

}
