/**
 * TemplateContext.java
 * Created on 2005/06/10
 * Copyright (c) by es , All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND 
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF 
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package izayoi.view;

import izayoi.Izayoi;
import izayoi.Util;
import izayoi.servlet.ServletConstants;

import java.io.IOException;
import java.nio.charset.Charset;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;

/**
 * ev[gReLXg
 * @author es
 */
public class TemplateContext {
    private Template template;
    private HttpServletRequest req;
    private HttpServletResponse res;
    private VelocityContext ctx;

    /**
     * @param path ev[g
     * @param request
     * @param response
     * @param encode ʂɏo͂ۂ̕R[h(ev[gt@C̎ۂ̕R[hƂ킹邱)
     */
    public TemplateContext(String path, HttpServletRequest request, HttpServletResponse response, Charset encode) {
        try {
            template = Velocity.getTemplate(path, encode.name());
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
        req = request;
        res = response;
        ctx = new VelocityContext();
        ctx.put("version", Izayoi.VERSION);
        ctx.put(ServletConstants.PARAM_STYLE, (String) req.getAttribute(ServletConstants.PARAM_STYLE));
        ctx.put(ServletConstants.PARAM_CONTEXT_ROOT, req.getContextPath());
    }

    /**
     * @param path ev[g
     * @param request
     * @param response
     */
    public TemplateContext(String path, HttpServletRequest request, HttpServletResponse response) {
        this(path, request, response, Util.utf8Charset);
    }

    /**
     * Recǉ
     * @param key
     * @param value
     */
    public void put(String key, Object value) {
        if (value != null) {
            ctx.put(key, value);
        }
    }

    /**
     * Recǉ
     * @param key
     * @param value
     */
    public void put(String key, Object[] value) {
        if (value != null && value.length > 0) {
            ctx.put(key, value);
        }
    }

    /**
     * R[h̃f[^Reiǉ
     * @param record
     */
    public void setRecord(DataContainer[] record) {
        put("records", record);
    }

    /**
     * uȂ܂ṽRecǉ
     * @param name
     */
    public void setName(String name) {
        put("filename", name);
        put("encfilename", Util.urlEncode(name));
    }

    /**
     * ut@CTCYṽRecǉ
     * @param size
     */
    public void setSize(long size) {
        ctx.put("size", Util.parseByteString(size));
    }

    /**
     * o
     * @throws IOException
     * @throws ServletException
     */
    public void finish() throws IOException, ServletException {
        res.setContentType("text/html; charset=" + template.getEncoding());
        try {
            template.merge(ctx, res.getWriter());
        } catch (IOException e) {
            throw e;
        } catch (Exception e) {
            throw new ServletException(e);
        }
        res.flushBuffer();
    }
}
