/*
 * ClientThread.java
 *
 * Created on 2004/08/29, 16:45
 */

package net.sourceforge.shingetsu.client;

import java.io.File;
import java.lang.Thread;
import java.util.Date;
import java.util.Enumeration;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sourceforge.shingetsu.cache.Cache;
import net.sourceforge.shingetsu.cache.CacheManager;
import net.sourceforge.shingetsu.cache.NetworkCacheAgent;
import net.sourceforge.shingetsu.node.Node;
import net.sourceforge.shingetsu.node.NodeList;
import net.sourceforge.shingetsu.util.Config;
import net.sourceforge.shingetsu.util.Time;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.lang.InterruptedException;
import java.lang.NumberFormatException;
import java.lang.NullPointerException;
import java.net.MalformedURLException;
import java.security.NoSuchAlgorithmException;

/**
 * m[h̕ێALbVt@C̍XVs߂̃NCAgXbh
 * @author Fuktommy
 * @author stokesia
 */
public class ClientThread extends Thread {
    
    private Config config = Config.getInstance();
    
    private long pingCycle = config.getPingCycle() * 60L;
    private long syncCycle = config.getSyncCycle() * 60L;
    private long initCycle = config.getInitCycle() * 60L;
    
    private long lastPingTime = 0L;
    private long lastSyncTime = 0L;
    private long lastInitTime = 0L;
    
    private long time = 0L;
    
    private long CLINENT_SLEEP_TIME = config.getClientSleepTime() * 60L * 1000L;
    private long START_SLEEP_TIME = 60L * 2L * 1000L;
    
    private NodeList nodeList = NodeList.getInstance();
    
    private CacheManager cacheManager = null;
    private NetworkCacheAgent networkCacheAgent = null;

    /** Creates a new instance of ClientThread */
    public ClientThread() {
        setPriority(Thread.MIN_PRIORITY);
        
        cacheManager = new CacheManager();
        networkCacheAgent = new NetworkCacheAgent();
    }

    /**
     * NCAgXbhs܂B
     */    
    public void run() {
        initialize();
        
        try {
            sleep(START_SLEEP_TIME);
        } catch(InterruptedException e) {
            System.err.println("Xbh̃X[v̊荞݂Ɏs܂B");
            e.printStackTrace();
        }
        
        while(true) {
            time = Time.getCurrentTime();
            
            if((time - lastPingTime) >= pingCycle) {
                ping();
                lastPingTime = Time.getCurrentTime();
            }
            
            time = Time.getCurrentTime();
            
            if((time - lastSyncTime) >= syncCycle) {
                synchronize();
                lastSyncTime = Time.getCurrentTime();
            }
            
            time = Time.getCurrentTime();
            
            if((time - lastInitTime) >= config.getInitCycle()) {
                initialize();
                lastInitTime = Time.getCurrentTime();
            }
            
            try {
                sleep(START_SLEEP_TIME);
            } catch(InterruptedException e) {
                System.err.println("Xbh̃X[v̊荞݂Ɏs܂B");
                e.printStackTrace();
            }
        }
        
    }
    
    private void initialize() {
        Node[] initNodes = config.getInitNodes();
        
        for(int i = 0; i < initNodes.length; i++) {
            Node node = initNodes[i];
        
                try {
                    if(node.ping()) {
                        Node responsedNode = node.join();
                        
                        if(!responsedNode.equals(node) && !nodeList.contains(responsedNode)) {
                            nodeList.add(responsedNode);
                        }
                    }
                } catch(MalformedURLException e) {
                    System.err.println("m[h " + node.toString() + " URLsłB");
                    e.printStackTrace();
                } catch(IOException e) {
                    String message = "m[h " + node.toString() + " Ƃ̒ʐMɓo̓G[܂B";
                    System.err.println(message);
                    e.printStackTrace();
                }
        }
        
        Enumeration enumeration = nodeList.enumeration();
        
        while(enumeration.hasMoreElements()) {
            Node node = (Node) enumeration.nextElement();
            
                try {
                    if(node.ping()) {
                        Node responsedNode = node.join();
                        
                        if(!responsedNode.equals(node) && !nodeList.contains(responsedNode)) {
                            nodeList.add(responsedNode);
                        }
                    }
                } catch(MalformedURLException e) {
                    System.err.println("m[h " + node.toString() + " URLsłB");
                    e.printStackTrace();
                } catch(IOException e) {
                    String message = "m[h " + node.toString() + " Ƃ̒ʐMɓo̓G[܂B";
                    System.err.println(message);
                    e.printStackTrace();
                }
        }
        
        for(int count = 0; count < config.getRetrySearchMax(); count++) {
            Node[] nodes = nodeList.getAll();
            
            for(int i = 0; i < nodes.length; i++) {
                Node node = nodes[i];
                
                try {
                    if(nodeList.size() <= (config.getNodeSize() - 1)) {
                        if(node.ping()) {
                            Node responsedNode = node.node();
                            
                            if(!nodeList.contains(responsedNode)) {
                                nodeList.add(responsedNode);
                            }
                        }
                    } else {
                        break;
                    }
                } catch(MalformedURLException e) {
                    System.err.println("m[h " + node.toString() + " URLsłB");
                    e.printStackTrace();
                } catch(IOException e) {
                    String message = "m[h " + node.toString() + " Ƃ̒ʐMɓo̓G[܂B";
                    System.err.println(message);
                    e.printStackTrace();
                }
            }
            
            if(nodeList.size() > (config.getNodeSize() - 1)) {
                break;
            }
        }
        
        nodeList.synchronize();
    }
    
    private void synchronize() {
        Enumeration enumeration = cacheManager.enumeration();
        
        while(enumeration.hasMoreElements()) {
            Cache cache = (Cache) enumeration.nextElement();
            File cacheFile = cache.getFile();
            
            String lastRecord = null;
            
            try {
                lastRecord = getLastRecord(cache);
            } catch(Exception e) {
                e.printStackTrace();
                continue;
            }
            
            String timeStamp = "0";
            
            /*
            String timeStamp = null;
            
            if(lastRecord != null) {
                timeStamp = getTimeStamp(lastRecord);
            } else {
                timeStamp = "0";
            }
             */
            
            String[] tokens = cacheFile.getName().split("\\.");
            String filename = tokens[0];
            String timeArgument = timeStamp + "-";
            
            String[] records = null;
            
            for(int i = 0; i < config.getRetrySearchMax(); i++) {
                records = networkCacheAgent.getRecordsWithHEAD(filename, timeArgument);
                
                if(records != null) {
                    break;
                }
            }
            
            if(((records != null) && (records.length == 0)) || (records == null)) {
                continue;
            } else {
                try {
                    records = cacheManager.difference(cache, records);
                } catch(FileNotFoundException e) {
                    System.out.println("LbVt@C " + cacheFile.getName() + " ݂܂B");
                    e.printStackTrace();
                    continue;
                }
            }
            
            if((records != null) && (records.length > 0)) {
                String record = records[0];
                
                tokens = record.split("<>");
                timeStamp = tokens[0];
                
                timeArgument =timeStamp + "-";
                
                for(int i = 0; i < config.getRetryGetMax(); i++) {
                    records = networkCacheAgent.getRecordsWithGET(filename, timeArgument);
                    
                    if(records != null) {
                        break;
                    }
                }
            }
            
            /*
            for(int i = 0; i < records.length; i++) {
                try {
                    records[i] = getRecord(filename, records[i]);
                } catch(Exception e) {
                    e.printStackTrace();
                    continue;
                }
            }*/
            
            if((records != null) && (records.length > 0)) {
                try {
                    cacheManager.synchronize(cache, records);
                } catch(FileNotFoundException e) {
                    System.out.println("LbVt@C " + cacheFile.getName() + " ݂܂B");
                } catch(NoSuchAlgorithmException e) {
                    System.out.println("w肳ꂽASY݂͑܂B");
                } catch(UnsupportedEncodingException e) {
                    System.out.println("w肳ĂGR[fBO͖łB");
                } catch(IOException e) {
                    System.out.println("o͒ɃG[܂B");
                }
            }
            
            try {
                cache.close();
            } catch(IOException e) {
                System.out.println("`l̃N[YɎs܂B");
            }
            
        }
        
    }

    private String getTimeStamp(String record) {
        String regex = "^(\\d+)<>";
        Matcher matcher = Pattern.compile(regex).matcher(record);
        
        assert matcher.find();
        
        String timeStamp = matcher.group();
        
        return timeStamp;
    }
    
    /*
    private String getRecord(String filename, String record) throws Exception {
        String[] tokens = record.split("<>");
        
        String timeStamp = tokens[0];
        String id = tokens[1];
        
        long time = 0L;
        
        try {
            time = Long.parseLong(timeStamp);
        } catch(NumberFormatException e) {
            System.out.println("t@C " + filename + " ̎ʎq " + id + " ^CX^vsłB");
            throw e;
        }
        
        for(int i = 0; i < config.getRetryGetMax(); i++) {
            record = networkCacheAgent.getRecord(filename, time, id);
            
            if(record != null) {
                break;
            }
        }
        
        return record;
    }
     */
    
    private String getLastRecord(Cache cache) throws Exception {
        String lastRecord = null;
        File cacheFile = cache.getFile();
        
        try {
            lastRecord = cache.getLastRecord();
        } catch(FileNotFoundException e) {
            System.out.println("LbVt@C " + cacheFile.getName() + " ݂܂B");
            throw e;
        } catch(IOException e) {
            System.out.println("LbVt@C̓o͒ɃG[܂B");
            System.out.println("LbVt@C: " + cacheFile.getName());
            throw e;
        }
        
        return lastRecord;
    }

    private void ping() {
        nodeList.pingAll();
        nodeList.synchronize();
    }

}
