/*
 * HttpRequestProcessor.java
 *
 * Created on 2004/08/19, 21:00
 */

package net.sourceforge.shingetsu.server;

import java.io.InputStream;
import java.io.FilterInputStream;
import java.io.ByteArrayOutputStream;
import java.io.BufferedInputStream;
//import java.io.PushbackInputStream;
import java.net.URL;
import java.net.URLDecoder;
import java.net.InetAddress;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

import net.sourceforge.shingetsu.server.HttpServer;
import net.sourceforge.shingetsu.server.HttpRequest;
import net.sourceforge.shingetsu.server.HttpRequestImpl;
import net.sourceforge.shingetsu.util.Config;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.lang.NumberFormatException;
import java.net.MalformedURLException;
import java.net.UnknownHostException;

import net.sourceforge.shingetsu.server.HttpProtocolException;

/**
 * @HTTP̃NGXg͂ɂĕKvƂȂ郁\bh񋟂܂B{I
 * HTTP̎dlɑď܂AׂĂ̎dlT|[g܂B
 * @author stokesia
 */
public class HttpRequestProcessor {
    
    private Config config = Config.getInstance();
    private String readingEncoding = "ISO-8859-1";
    
    private final int bufferSize = config.getRequestBufferSize();
    //private int readBufferSize = bufferSize;
    
    /** Creates a new instance of HttpRequestProcessor */
    private HttpRequestProcessor() {
    }
    
    /**
     * NGXgvZbT̃CX^X擾܂B
     * @return NGXgvZbT̃CX^X
     */    
    public static HttpRequestProcessor getProcessor() {
        return new HttpRequestProcessor();
    }
    
    /**
     * ̓Xg[HTTPɂĉ͂܂B
     * @param in ̓Xg[
     * @param address ڑzXg̃AhX
     * @throws IOException ͒ɃG[ꍇ
     * @throws HttpProtocolException HTTPNGXg͒ɃG[ꍇ
     * @return ̃NGXgHttpRequestIuWFNg
     */    
    public HttpRequest parseRequest(InputStream in, InetAddress address) 
        throws IOException, HttpProtocolException {

        /*    
        PushbackInputStream inputStream = new PushbackInputStream(
                                                                                    new BufferedInputStream(in), bufferSize);
         */
            
        in = new BufferedInputStream(in, bufferSize);
         
        HttpRequestImpl request = new HttpRequestImpl();
        
        request = readRequestLine(request, in);
        request = readMessageHeader(request, in);
        request = readMessageBody(request, in);
        
        //request = readRequestLine(request, inputStream);
        //request = readMessageHeader(request, inputStream);
        //request = readMessageBody(request, inputStream);
        
        //pushbackInputStream.close();
        
        request.setInetAddress(address);
        
        return request;
    }
    
    //private HttpRequestImpl readRequestLine(HttpRequestImpl request, PushbackInputStream in) 
    private HttpRequestImpl readRequestLine(HttpRequestImpl request, InputStream in) 
        throws IOException, HttpProtocolException {
        String line = readLine(in);
        
        line = replaceCRLF(line);
        String[] tokens = parseRequestLine(line);

        String method = tokens[0];
        String originalRequestURI = tokens[1];
        String httpVersion = tokens[2];
        
        URL url = createURL(originalRequestURI);
        
        request.setMethod(method);
        request.setProtocolVersion(httpVersion);
        request.setURL(url);
        
        readLine(in); /*    󔒍s̓ǂݍ    */
        
        return request;
    }
    
    //private HttpRequestImpl readMessageHeader(HttpRequestImpl request, PushbackInputStream in) 
private HttpRequestImpl readMessageHeader(HttpRequestImpl request, InputStream in) 
        throws IOException, HttpProtocolException {
        Hashtable messageHeaderTable = new Hashtable();

        String headerRegex = "[-_a-zA-Z]+\\s*:\\s*.+";
        Pattern headerPattern = Pattern.compile(headerRegex, Pattern.DOTALL);
        
        /*  {͌߂͗ǂȂ    */
        String multipartRegex = "Content-Type:\\s+multipart/form-data;\\s+boundary=(.+)";
        Pattern multipartPattern = Pattern.compile(multipartRegex, Pattern.DOTALL);
        String boundary = null;

        boolean startBoundary = false;
        boolean endBoundary = false;
        boolean multipart = false;
        
        Vector partList = new Vector();
        
        String headerLine = null;
        
        while((headerLine = readLine(in)) != null) {
            if(multipartPattern.matcher(headerLine).matches()) {
                multipart = true;
                request.setMultipatRequest(multipart);
                boundary = detectBoundary(headerLine);
            }
            
            String[] header = createHeaderValue(headerLine);
            messageHeaderTable.put(header[0], header[1]);   /*  0:key 1:value   */
            
            if(multipart) {
                break;
            }
        }
            
        if(!multipart) {
            request.setMessageHeaderTable(messageHeaderTable);
            return request;
        }
        
        while((headerLine = readLineCRLF(in)) != null) {
            if(endBoundary && (headerLine.length() == 0)) {
                break;
            }
            
            if(!startBoundary && headerLine.equals("--" + boundary + HttpServer.LINE_FEED_CRLF)) {
                startBoundary = true;
            } else if(headerLine.equals("--" + boundary + HttpServer.LINE_FEED_CRLF)
            || headerLine.equals("--" + boundary + "--" + HttpServer.LINE_FEED_CRLF)) {
                Object[] header = createPartValue(partList);
                messageHeaderTable.put(header[0], header[1]);   /*  0:key 1:value   */
                
                partList.clear();
                
                if(headerLine.equals("--" + boundary + "--" + HttpServer.LINE_FEED_CRLF)) {
                    endBoundary = true;
                    break;
                }
            } else {
                if(startBoundary) {
                    partList.add(headerLine);
                }
            }
            
        }

        request.setMessageHeaderTable(messageHeaderTable);
        
        return request;
    }
    
    //private HttpRequestImpl readMessageBody(HttpRequestImpl request, PushbackInputStream in) 
    private HttpRequestImpl readMessageBody(HttpRequestImpl request, InputStream in) 
        throws IOException, HttpProtocolException {
        String method = request.getMethod();
        
        String query = null;
            
        if(method.equals(HttpServer.METHOD_POST) && !request.isMultipartRequest()) {
            String contentLength = request.getParameter("Content-Length");
            int length = 0;
            
            try {
                length = Integer.parseInt(contentLength);
            } catch(NumberFormatException e) {
                String message = "wb_̍\sA܂͂̃g[NsłB";
                throw new HttpProtocolException(message, HttpServer.BAD_REQUEST);
            }
            
            query = readLine(in, length);
        } else {
            query = request.getQuery();
        }
        
        Hashtable queryParameterTable = createQueryParameterTable(query);
        request.setQueryParameterTable(queryParameterTable);

        return request;
    }
    
    private Hashtable createQueryParameterTable(String query) throws HttpProtocolException {
        Hashtable queryParameterTable = null;
        
        if(query != null) {
            queryParameterTable = new Hashtable();
            
            String[] tokens = query.split("&");
            
            for(int i = 0; i < tokens.length; i++) {
                String[] elements = tokens[i].split("=");
                
                if(elements.length == 2) {
                    String name = elements[0];
                    String value = elements[1];
                    
                    try {
                        value = URLDecoder.decode(value, config.getGeneralEncoding());
                    }  catch(UnsupportedEncodingException e) {
                        String message = "NGXgNG[̃fR[fBOɎs܂B";
                        throw new HttpProtocolException(message, HttpServer.INTERNAL_SERVER_ERROR);
                    }
                    
                    queryParameterTable.put(name, value);
                }
            }
        }
        
        return queryParameterTable;
    }
    
    private String readLine(InputStream in) throws IOException {
        String line = readLineCRLF(in);

        if(line != null) {
            line  = replaceCRLF(line);
            
            if(line.length() == 0) {
                line = null;
            }
        }

        return line;
    }

    private String readLine(InputStream in, int length) throws IOException {
        byte[] buffer = new byte[length];
        ByteArrayOutputStream byteArray = new ByteArrayOutputStream();
        
        in.read(buffer);
        byteArray.write(buffer);
        
        String line = byteArray.toString(readingEncoding);
        
        if(line.length() == 0) {
            line = null;
        }

        return line;
    }
    
    //private String readLine(PushbackInputStream in) throws IOException {
    private String readLineCRLF(InputStream in) throws IOException {
        byte[] buffer = new byte[1];
        ByteArrayOutputStream byteArray = new ByteArrayOutputStream();
        
        while(in.read(buffer) > 0) {
            byteArray.write(buffer);
            if(buffer[0] == '\n') {
                break;
            }
        }
        
        String line = byteArray.toString(readingEncoding);
        
        if(line.length() == 0) {
            line = null;
        }

        return line;
    }
    
    /*
    private String readLine(PushbackInputStream in) throws IOException {
        byte[] buffer = new byte[bufferSize];
        ByteArrayOutputStream byteArray = new ByteArrayOutputStream();

        int size = 0;
        boolean end = false;
        
        FilterInputStream filterIn = in;
        readBufferSize = checkAvailableBuffer(in);
        
        while((readBufferSize != 0) && ((size = filterIn.read(buffer, 0, readBufferSize)) > 0)) {
            for(int i = 0; i < size; i++) {
                byteArray.write(buffer, i, 1);
                
                if(buffer[i] == '\n' ) {
                    end = true;
                    
                    if((i + 1) <= buffer.length) {
                        readBufferSize = size - (i + 1);
                        in.unread(buffer,i + 1, readBufferSize);
                    }
                    break;
                }
            }
            
            if(end) {
                break;
            }
            
            readBufferSize = checkAvailableBuffer(in);
        }
        
        String line = byteArray.toString(readingEncoding);
        
        if(line.length() == 0) {
            line = null;
        }

        return line;
    }
     */
    
    /*
    private int checkAvailableBuffer(InputStream in) throws IOException {
        readBufferSize = in.available();
        
        if(readBufferSize > bufferSize) {
            readBufferSize = bufferSize;
        }
        
        return readBufferSize;
    }
     */
    
    private String[] parseRequestLine(String line) throws HttpProtocolException {
        /*  regex pattern = \s+ */
        String[] tokens = line.split("\\s+");
        
        if(!(tokens.length > 2)) {
            String message = "NGXgC̍\sA܂͂̃g[NsłB";
            throw new HttpProtocolException(message, HttpServer.BAD_REQUEST);
        }

        /*  regex pattern = HTTP/0.9|HTTP/1.0|HTTP/1.1    */
        String versionRegex = "(" + HttpServer.HTTP_VERSION_0_9 + "|" 
                                                         + HttpServer.HTTP_VERSION_1_0 + "|"
                                                         + HttpServer.HTTP_VERSION_1_1 + ")";
        
        String commandRegex = "(" + HttpServer.METHOD_GET + "|" 
                                                              + HttpServer.METHOD_POST + ")";

        if(!(tokens.length == 3)
                    || !(tokens[0].matches(commandRegex))
                    || !(tokens[1].toLowerCase().matches("(^http://.+)|(^/.*)"))
                    || !(tokens[2].matches(versionRegex))) {
            String message = "NGXgC̍\sA܂͂̃g[NsłB";
            throw new HttpProtocolException(message, HttpServer.BAD_REQUEST);
        }

        return tokens;
    }
    
    private String detectBoundary(String headerLine) {
        int boundaryGroup = 1;
        String regex = "Content-Type:\\s+multipart/form-data;\\s+boundary=(.+)";
        
        Pattern contentTypePattern = Pattern.compile(regex);
        Matcher matcher = contentTypePattern.matcher(headerLine);
        
        assert matcher.find();
        
        String boundary = matcher.group(boundaryGroup);
        
        return boundary;
    }
    
    private String detecteHeaderName(String headerValue) {
        int nameGroup = 1;
        
        String regex = "name=\"([^\"]+)\"";
        Matcher matcher = Pattern.compile(regex).matcher(headerValue);
        
        assert matcher.find();
        
        String name = matcher.group(nameGroup);
        
        return name;
    }

    private String[] createHeaderValue(String headerLine) {
        String regex = "([-_a-zA-Z]+)\\s*:\\s*(.+)";
    
        Matcher matcher = Pattern.compile(regex, Pattern.DOTALL).matcher(headerLine);
        assert matcher.matches();
        
        int nameGroup = 1;
        int valueGroup = 2;
        
        String[] header = new String[2];
        
        header[0] = matcher.group(nameGroup);
        header[1] = replaceCRLF(matcher.group(valueGroup));
        
        return header;
    }

    private URL createURL(String requestURI) throws HttpProtocolException {
        String hostAddress = null;
        
        try {
            hostAddress = InetAddress.getLocalHost().getHostAddress();
        } catch(UnknownHostException e) {
            String message = "[JzXgIPAhX擾邱Ƃł܂łB";
            throw new HttpProtocolException(message, HttpServer.INTERNAL_SERVER_ERROR);
        }
        
        if(requestURI.matches("^/.*")) {    /*  regex pattern = ^/.* */
            requestURI = "http://" + hostAddress + requestURI;
        } 
        
        URL url = null;
        
        try {
            url = new URL(requestURI);
        } catch(MalformedURLException e) {
            String message = "NGXgC̃NGXgURI̍\słB";
            throw new HttpProtocolException(message, HttpServer.BAD_REQUEST);
        }
        
        return url;
    }
    
    private Object[] createPartValue(Vector multipartList) throws HttpProtocolException {
        Hashtable headerTable = new Hashtable();
        Enumeration enumeration = multipartList.elements();
        
        boolean startMultipart = false;
        boolean attachment = false;
        
        String name = null;
        StringBuffer multipart = new StringBuffer();
        
        while(enumeration.hasMoreElements()) {
            String headerLine = (String) enumeration.nextElement();
            
            if(!startMultipart) {
                headerLine = replaceCRLF(headerLine);
                    
                if(headerLine.matches("[-_a-zA-Z]+\\s*:\\s*.+")) {
                    String[] header = createHeaderValue(headerLine);
                    headerTable.put(header[0], header[1]);   /*  0:key 1:value   */
                    
                    if(header[0].equals("Content-Disposition")) {
                        name = detecteHeaderName(header[1]);
                    }
                }
                
                if(containsAttachment(headerLine)) {
                    attachment = true;
                }
                
                if(headerLine.length() == 0) {
                    startMultipart = true;
                }
            } else {
                multipart.append(headerLine);
            }
        }
        
        name = convertTo(name);
        
        String value = replaceCRLF(multipart.toString());
        
        if(!attachment) {
            value = convertTo(value);
        }

        headerTable.put(name, value);
        
        Object[] header = new Object[2];

        header[0] = name;
        header[1] = headerTable;
        
        return header;
    }
    
    private boolean containsAttachment(String headerLine) {
        Matcher matcher = Pattern.compile("filename=\"[^\"]+\"").matcher(headerLine);
        
        boolean result = matcher.find();
        
        return result;
    }
    
    private String convertTo(String string) throws HttpProtocolException {
        String encoding = config.getGeneralEncoding();
        
        try {
            byte[] bytes = string.getBytes(readingEncoding);
            string = new String(bytes, encoding);
        } catch(UnsupportedEncodingException e) {
            String message = "w肳ꂽGR[fBO̓T|[gĂ܂B";
            throw new HttpProtocolException(message, HttpServer.INTERNAL_SERVER_ERROR);
        }
        
        return string;
    }
    
    private String replaceCRLF(String line) {
        return line.replaceAll("\\r?\\n$", "");
    }

}
