/*
 * ServerCGI.java
 *
 * Created on 2004/08/21, 16:20
 */

package net.sourceforge.shingetsu.server.cgi;

import java.io.BufferedReader;
import java.io.PrintWriter;
import java.io.File;
import java.io.InputStreamReader;
import java.io.FileInputStream;
import java.lang.Math;
import java.lang.Long;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.util.Enumeration;
import java.util.Vector;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

import net.sourceforge.shingetsu.cache.Cache;
import net.sourceforge.shingetsu.node.Node;
import net.sourceforge.shingetsu.node.NodeList;
import net.sourceforge.shingetsu.server.HttpRequest;
import net.sourceforge.shingetsu.server.HttpResponse;
import net.sourceforge.shingetsu.server.HttpServer;
import net.sourceforge.shingetsu.server.cgi.PlugIn;
import net.sourceforge.shingetsu.util.Config;

import java.lang.NumberFormatException;
import java.lang.IllegalArgumentException;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.net.MalformedURLException;


import net.sourceforge.shingetsu.server.HttpProtocolException;

/**
 * ̃m[h̃T[og̃R}hserver.cgiɑNXłB
 * @author stokesia
 */
public final class ServerCGI extends PlugIn {
    
    private Config config = Config.getInstance();
    
    private NodeList nodeList = NodeList.getInstance();
    
    /*  regex pattern = ^/server.cgi/?    */
    private String rootPattern = "^/server.cgi/?";
    
    /*  regex pattern = ^/server.cgi/ping    */
    private String pingPattern = "^/server.cgi/ping";
    
    /*  regex pattern = ^/server.cgi/node    */
    private String nodePattern = "^/server.cgi/node";
    
    /*  regex pattern = ^/server.cgi/join/([^:]*:(\d+).*)    */
    private String joinPattern = "^/server.cgi/join/([^:]*:(\\d+).*)";
    
    /*  regex pattern = ^/server.cgi/bye/([^:]*:(\d+).*)    */
    private String byePattern = "^/server.cgi/bye/([^:]*:(\\d+).*)";
    
    /*  regex pattern = ^/server.cgi/have/(\w+)    */
    private String havePattern = "^/server.cgi/have/(\\w+)";
    
/*  regex pattern = ^/server.cgi/get/(\w+)/(-?\d*-?\d*-?/?[0-9a-f]*)   */
    private String getPattern = "^/server.cgi/get/(\\w+)/(-?\\d*-?\\d*-?/?[0-9a-f]*)";
    
    /*  regex pattern = ^/server.cgi/head/(\w+)/(-?\d*-?\d*-?/?[0-9a-f]*)   */
    private String headPattern = "^/server.cgi/head/(\\w+)/(-?\\d*-?\\d*-?/?[0-9a-f]*)";
    
    /*  regex pattern = ^/server.cgi/update/(\w+)/(\d+)/([0-9a-f]+)/([^:]*:(\d+).*)  */
    private String updatePattern = "^/server.cgi/update/(\\w+)/(\\d+)/([0-9a-f]+)/([^:]*:(\\d+).*)";

    /** Creates a new instance of ServerCGI */
    public ServerCGI() {
    }
    
    public void excecute(HttpRequest request, HttpResponse response) 
        throws HttpProtocolException, IOException {
        String path = request.getPath();

        if(path.matches(rootPattern)) {
            root(request, response);
        } else if(path.matches(pingPattern)) {
            ping(request, response);
        } else if(path.matches(nodePattern)) {
            node(request, response);
        } else if(path.matches(joinPattern)) {
            join(request, response);
        } else if(path.matches(byePattern)) {
            bye(request, response);
        } else if(path.matches(havePattern)) {
            have(request, response);
        } else if(path.matches(getPattern)) {
            get(request, response);
        } else if(path.matches(headPattern)) {
            head(request, response);
        } else if(path.matches(updatePattern)) {
            update(request, response);
        } else {
            String message = "vꂽURL " + request.getPath() + " ͌܂łB";
            throw new HttpProtocolException(message, HttpServer.NOT_FOUND);
        }

    }
    
    private void root(HttpRequest request, HttpResponse response) 
        throws HttpProtocolException, IOException {
        String encoding = "EUC-JP";
        CharsetDecoder decoder = Charset.forName(encoding).newDecoder();
        
        String motdFileName = config.getMotdFilename() ;
        String motdFilePath = config.getDocumentRootDirectory() + File.separator 
                                                        + motdFileName;
        
        BufferedReader reader = null;
        
        try {
            reader = new BufferedReader(new InputStreamReader(
                                    new FileInputStream(motdFilePath), decoder));
        } catch(FileNotFoundException e) {
            String message = "vꂽURL " + request.getPath() + " ͌܂łB";
            throw new HttpProtocolException(message, HttpServer.NOT_FOUND);
        } catch(IOException e) {
            String message = "t@C " + motdFileName +"̓o͒ɃG[܂B";
            throw new HttpProtocolException(message, HttpServer.INTERNAL_SERVER_ERROR);
        }
        
        response.setContentType("text/plain; charset=" + config.getGeneralEncoding());
        
        PrintWriter writer = response.getWriter();
        
        String line = null;
        
        while((line = reader.readLine()) != null) {
            writer.println(line);
        }
        
        reader.close();
        
        writer.flush();
        writer.close();
    }

    private void ping(HttpRequest request, HttpResponse response) 
        throws HttpProtocolException, IOException {
        String address = request.getHostAddress();
        
        response.setContentType("text/plain; charset=" + config.getGeneralEncoding());
        
        PrintWriter writer = response.getWriter();
        
        writer.println("PONG");
        writer.println(address);
        
        writer.flush();
        writer.close();
    }

    private void node(HttpRequest request, HttpResponse response) 
        throws HttpProtocolException, IOException {
        Node node = nodeList.getRandom();
        
        response.setContentType("text/plain; charset=" + config.getGeneralEncoding());
        
        PrintWriter writer = response.getWriter();
        
        writer.println(node.toString());
        
        writer.flush();
        writer.close();
    }
    
    private void join(HttpRequest request, HttpResponse response) 
        throws HttpProtocolException, IOException {
        String path = request.getPath();
        
        Matcher matcher = Pattern.compile(joinPattern).matcher(path);
        
        assert matcher.matches();
        
        int nodeAddressGroup = 1;
        String nodeAddress = matcher.group(nodeAddressGroup).replace('+', '/');

        /*  regex pattern = ^:(\d+)/.*    */
        if(nodeAddress.matches("^:(\\d+)/.*")) {
            nodeAddress = request.getRemoteHost()+ nodeAddress;
        }
        
        Node node = null;
        
        try {
            node = new Node(nodeAddress);
        } catch(MalformedURLException e) {
            String message = "w肳ꂽm[hAhX"+ nodeAddress + "Lł͂܂łB";
            throw new HttpProtocolException(message, HttpServer.INTERNAL_SERVER_ERROR);
        } 
        
        if(!node.ping()) {
            String message = "w肳ꂽm[h"+ nodeAddress + "Lł͂܂łB";
            throw new HttpProtocolException(message, HttpServer.INTERNAL_SERVER_ERROR);
        }

        Node responsedNode = nodeList.getRandom();

        response.setContentType("text/plain; charset=" + config.getGeneralEncoding());
        
        PrintWriter writer = response.getWriter();
        
        writer.println("WELCOME");
        
        if(Math.random() < 0.5) {
            writer.println(responsedNode.toString());
        }
        
        nodeList.add(node);
        nodeList.synchronize();
        
        writer.flush();
        writer.close();
    }
    
    private void bye(HttpRequest request, HttpResponse response) 
        throws HttpProtocolException, IOException {
        String path = request.getPath();
       
        Matcher matcher = Pattern.compile(byePattern).matcher(path);
        
        assert matcher.matches();
        
        int nodeAddressGroup = 1;
        String nodeAddress = matcher.group(nodeAddressGroup).replace('+', '/');
        
        /*  regex pattern = ^:(\d+)/.*    */
        if(nodeAddress.matches("^:(\\d+)/.*")) {
            nodeAddress = request.getRemoteHost()+ nodeAddress;
        }
        
        Node node = null;
        
        try {
            node = new Node(nodeAddress);
        } catch(MalformedURLException e) {
            String message = "w肳ꂽm[hAhX"+ nodeAddress + "Lł͂܂łB";
            throw new HttpProtocolException(message, HttpServer.INTERNAL_SERVER_ERROR);
        } 
        
        if(!node.ping()) {
            String message = "w肳ꂽm[h"+ nodeAddress + "Lł͂܂łB";
            throw new HttpProtocolException(message, HttpServer.INTERNAL_SERVER_ERROR);
        }
        
        nodeList.remove(node);
        nodeList.synchronize();

        response.setContentType("text/plain; charset=" + config.getGeneralEncoding());
        
        PrintWriter writer = response.getWriter();
        
        writer.println("BYEBYE");

        writer.flush();
        writer.close();
    }
    
    private void have(HttpRequest request, HttpResponse response) 
        throws HttpProtocolException, IOException {
        String path = request.getPath();

        Matcher matcher = Pattern.compile(havePattern).matcher(path);
        
        assert matcher.matches();
        
        int filenameGroup = 1;
        
        String filename = matcher.group(filenameGroup) + "." + config.getDataFileExtension();
        String filePath = config.getDataDirectory() + File.separator + filename;
        
        File file = new File(filePath);
        
        response.setContentType("text/plain; charset=" + config.getGeneralEncoding());
        
        PrintWriter writer = response.getWriter();
        
        if(file.exists()) {
            writer.println("YES");
        } else {
            writer.println("NO");
        }
        
        writer.flush();
        writer.close();
    }
    
    private void get(HttpRequest request, HttpResponse response) 
        throws HttpProtocolException, IOException {
        String path = request.getPath();

        Matcher matcher = Pattern.compile(getPattern).matcher(path);
        
        assert matcher.matches();
        
        int filenameGroup = 1;
        int timeArgumentGroup = 2;
        
        String filename = matcher.group(filenameGroup);
        String timeArgument = matcher.group(timeArgumentGroup);
       
        String[] records = null;
        
        try {
            records = searchRecord(request, filename, timeArgument);
        } catch(NumberFormatException e) {
            String message ="vꂽpX " + path +  " ̎słB";
            throw new HttpProtocolException(message, HttpServer.INTERNAL_SERVER_ERROR);
        } catch(IOException e) {
            String message = "t@C̓o͒ɃG[܂B";
            throw new HttpProtocolException(message, HttpServer.INTERNAL_SERVER_ERROR);
        } catch(IllegalArgumentException e) {
            String message ="vꂽpX " + path +  " ̓t̏słB";
            throw new HttpProtocolException(message, HttpServer.INTERNAL_SERVER_ERROR);
        }
        
        response.setContentType("text/plain; charset=" + config.getGeneralEncoding());
        
        PrintWriter writer = response.getWriter();
        
        if(records != null) {
            for(int i = 0; i < records.length; i++) {
                if(!isRemoveRecord(records[i])) {
                    writer.println(records[i]);
                }
            }
        }
        
        writer.flush();
        writer.close();
    }
    
    private void head(HttpRequest request, HttpResponse response) 
        throws HttpProtocolException, IOException {
        String path = request.getPath();
        
        Matcher matcher = Pattern.compile(headPattern).matcher(path);
        
        assert matcher.matches();
        
        int filenameGroup = 1;
        int timeArgumentGroup = 2;
        
        String filename = matcher.group(filenameGroup);
        String timeArgument = matcher.group(timeArgumentGroup);

        String[] records = null;
        
        try {
            records = searchRecord(request, filename, timeArgument);
        } catch(NumberFormatException e) {
            String message ="vꂽpX " + path +  " ̎słB";
            throw new HttpProtocolException(message, HttpServer.INTERNAL_SERVER_ERROR);
        } catch(IOException e) {
            String message = "t@C̓o͒ɃG[܂B";
            throw new HttpProtocolException(message, HttpServer.INTERNAL_SERVER_ERROR);
        } catch(IllegalArgumentException e) {
            String message ="vꂽpX " + path +  " ̓t̏słB";
            throw new HttpProtocolException(message, HttpServer.INTERNAL_SERVER_ERROR);
        }
        
        response.setContentType("text/plain; charset=" + config.getGeneralEncoding());
        
        PrintWriter writer = response.getWriter();

        if(records != null) {
            for(int i = 0; i < records.length; i++) {
                if(!isRemoveRecord(records[i])){
                    String[] tokens = records[i].split("<>");
                    writer.println(tokens[0] + "<>" + tokens[1]);
                }
            }
        }

        writer.flush();
        writer.close();
    }
    
    private String[] searchRecord(HttpRequest request, String filename, String timeArgument) 
        throws IOException, HttpProtocolException {
        File file  = getCacheFile(filename);

        if(!(file.exists())) {
            String message = "vꂽURL " + request.getPath() + " ͌܂łB";
            throw new HttpProtocolException(message, HttpServer.NOT_FOUND);
        }

        Cache cache = null;
        
        try {
            cache = new Cache(file);
        } catch(Exception e) {
            String message = "vꂽURL " + request.getPath() + " ͌܂łB";
            throw new HttpProtocolException(message, HttpServer.NOT_FOUND);
        }
        
        String[] records = null;

        try {
            records = cache.searchRecord(timeArgument);
        } catch(NumberFormatException e) {
            String message = "vꂽURL " + request.getPath() + " ͌܂łB";
            throw new HttpProtocolException(message, HttpServer.NOT_FOUND);
        }
        
        return records;
    }
    
    private void update(HttpRequest request, HttpResponse response) 
        throws HttpProtocolException, IOException {
        String path = request.getPath();
        
        Matcher matcher = Pattern.compile(updatePattern).matcher(path);
        
        assert matcher.matches();
        
        int filenameGroup = 1;
        int timeStampGroup  = 2;
        int idGroup = 3;
        int nodeAddressGroup = 4;
        
        String filename = matcher.group(filenameGroup);
       
        long time = 0;

        try {
            time = Long.parseLong(matcher.group(timeStampGroup));
        } catch(NumberFormatException e) {
            String message = "vꂽURL " + request.getPath() + " ͌܂łB";
            throw new HttpProtocolException(message, HttpServer.NOT_FOUND);
        }
        
        response.setContentType("text/plain; charset=" + config.getGeneralEncoding());
        //response.setStatusCode(HttpServer.NOT_CONTENT);
            
        PrintWriter writer = response.getWriter();
            
        writer.flush();
        writer.close();

        String id = matcher.group(idGroup);
        String nodeAddress = matcher.group(nodeAddressGroup).replace('+', '/');
        
        /*  regex pattern = ^:(\d+)/.*    */
        if(nodeAddress.matches("^:(\\d+)/.*")) {
            nodeAddress = request.getRemoteHost() + nodeAddress;
        }
        
        try {
            nodeAddress = updateCache(filename, time, id, nodeAddress);
        } catch(Exception e) {
            e.printStackTrace();
        }
        
        try {
            informUpdate(nodeAddress, time, filename, id);
        } catch(Exception e) {
            e.printStackTrace();
        }

    }
    
    private String updateCache(String filename, long time, String md5, String nodeAddress)
        throws Exception {  
        File file = getCacheFile(filename);    
        
        if(file.exists()) {
            Cache cache = new Cache(file);
            
            if(!cache.containsRecord(time, md5)) {
                String timeArgument = new Long(time).toString() + "/" + md5;
                
                Node node = new Node(nodeAddress);
                
                String[] records = node.get(filename,  timeArgument);
                String record = records[0];
                
                int limit = 3;
                String[] tokens = record.split("<>", limit);
                
                String body = tokens[2];
                
                if(config.isCheckMD5() && md5.equals(getMD5(body))) {
                    cache.lock();
                    
                    cache.append(records[0]);
                    
                    cache.unlock();
                    cache.close();
                }

                nodeAddress = ":" + config.getPort() + "/server.cgi";
            }
        }
        
        return nodeAddress;
    }

}
